"""This module provides base implementations for HTML stripping.

The default implementation will use bleach if it is installed, otherwise an overly aggressive,
very simple html stripper is employed.
"""
import re
from typing import Callable, List, Mapping

import bleach

HtmlTags = List[str]
HtmlAttributes = Mapping[str, List[str]]

HtmlStripper = Callable[[HtmlTags, HtmlAttributes, str], str]


_html_re = re.compile(r"""<[^>]*?>""")


def naive_html_stripper(_: HtmlTags, __: HtmlAttributes, val: str) -> str:
    return _html_re.sub("", val)


def bleach_html_stripper(tags: HtmlTags, attributes: HtmlAttributes, val: str) -> str:
    """Exclude tags and attributes using BeautifulSoup4."""
    return bleach.clean(val, tags, attributes, strip=True, strip_comments=True)


def default_html_stripper(tags: HtmlTags, attributes: HtmlAttributes, val: str) -> str:
    return bleach_html_stripper(tags, attributes, val)



"""Contains the Messages data structure."""
from typing import Dict, List, Mapping

_GENERAL_SUBJECT = ""

MessageData = Mapping[str, List[str]]


class Messages:
    """Contains a mapping of subjects to a list of messages."""

    def __init__(self) -> None:
        self._data: Dict[str, List[str]] = {}

    def __bool__(self):
        return self.has_messages

    @property
    def has_messages(self) -> bool:
        return bool(self._data)

    @property
    def messages(self) -> MessageData:
        return self._data

    def add_message(self, subject: str, message: str) -> "Messages":
        if subject not in self._data:
            self._data[subject] = []
        self._data[subject].append(message)
        return self

    def add_general_message(self, message: str) -> "Messages":
        return self.add_message(_GENERAL_SUBJECT, message)

    def has_message_for(self, subject: str) -> bool:
        return subject in self._data

    def has_general_message(self) -> bool:
        return self.has_message_for(_GENERAL_SUBJECT)



"""Dict serialization methods."""
from base64 import b64encode
from datetime import date as Date
from datetime import datetime as DateTime
from decimal import Decimal
from typing import (Any, Callable, Iterable, List, Mapping, Optional, Sequence,
                    Set, TypeVar)
from uuid import UUID

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
Record = Mapping[str, Any]
Bytes = bytes


class Serialize:
    """Contains serialization methods. They're placed as static methods on a class to avoid
    name conflicts.
    """

    @staticmethod
    def map(fn: Callable[[T], U]) -> Callable[[Iterable[T]], List[U]]:
        """Returns a function that accepts a iterable and maps over it with fn."""

        def mapper(items: Iterable[T]) -> List[U]:
            return [fn(item) for item in items]

        return mapper

    @staticmethod
    def pick(keys: Sequence[str]) -> Callable[[Record], Record]:
        """Returns a new mapping with only the keys specified."""

        def picker(item: Record) -> Record:
            return {k: item[k] for k in keys if k in item}

        return picker

    @staticmethod
    def exclude(keys: Set[str]) -> Callable[[Record], Record]:
        """Returns a new mapping without the keys specified."""

        def excluder(item: Record) -> Record:
            return {k: v for (k, v) in item.items() if k not in keys}

        return excluder

    @staticmethod
    def evolve(maps: Mapping[str, Callable[[Any], Any]]) -> Callable[[Record], Record]:
        """Returns a new mapping with certain keys mapped as specified."""

        def evolver(item: Record) -> Record:
            return {k: (maps[k](v) if k in maps else v) for (k, v) in item.items()}

        return evolver

    @staticmethod
    def compose(outer: Callable[[U], V], inner: Callable[[T], U]) -> Callable[[T], V]:
        """Returns a single function that composes outer of inner."""

        def composer(val: T) -> V:
            return outer(inner(val))

        return composer

    @staticmethod
    def optional(fn: Callable[[T], U]) -> Callable[[Optional[T]], Optional[U]]:
        """Returns a function that will return None if given None, otherwise maps the value with
        fn."""

        def optionaler(val: Optional[T]) -> Optional[U]:
            if val is None:
                return None
            return fn(val)

        return optionaler

    @staticmethod
    def date(val: Date) -> str:
        """date to ISO format."""
        return val.isoformat()

    @staticmethod
    def datetime(val: DateTime) -> str:
        """datetime to ISO format."""
        return val.isoformat()

    @staticmethod
    def uuid(val: UUID) -> str:
        """UUID to string."""
        return str(val)

    @staticmethod
    def decimal(val: Decimal) -> str:
        """Decimal number to string."""
        return str(val)

    @staticmethod
    def bytes(val: Bytes) -> str:
        """Bytes to b64 string."""
        return b64encode(val).decode("utf-8")



"""Validator and Validations."""
import inspect
import re
from dataclasses import dataclass
from decimal import Decimal
from typing import (Any, Awaitable, Callable, List, Mapping, MutableMapping,
                    Optional, Sequence, Type, Union, cast)

from .validation_result import ValidationResult

Record = Mapping[str, Any]
MutableRecord = MutableMapping[str, Any]
ValidateFieldSync = Callable[[Any], bool]
ValidateFieldAsync = Callable[[Any], Awaitable[bool]]
ValidateField = Union[ValidateFieldSync, ValidateFieldAsync]
ValidateRecordSync = Callable[[Record], bool]
ValidateRecordAsync = Callable[[Record], Awaitable[bool]]
ValidateRecord = Union[ValidateRecordSync, ValidateRecordAsync]

NUMBER_TYPES = (int, float, Decimal)
Number = Union[int, float, Decimal]


@dataclass
class FieldValidator:
    """Information to validate a single field."""

    field_name: str
    message: str
    validator: ValidateField

    async def validate(self, value: Any) -> bool:
        """Properly calls the validator."""
        if inspect.iscoroutinefunction(self.validator):
            return await cast(ValidateFieldAsync, self.validator)(value)
        return cast(ValidateFieldSync, self.validator)(value)


@dataclass
class RecordValidator:
    """Information to validate a whole record."""

    message: str
    validator: ValidateRecord
    field_name: Optional[str]

    async def validate(self, record: Record) -> bool:
        """Properly calls the validator."""
        if inspect.iscoroutinefunction(self.validator):
            return await cast(ValidateRecordAsync, self.validator)(record)
        return cast(ValidateRecordSync, self.validator)(record)


ValidatorType = Union[FieldValidator, RecordValidator]


class Validator:
    """Records validators for a record."""

    def __init__(self, validators: Optional[List[ValidatorType]] = None):
        self.validators: List[ValidatorType] = validators or []

    @staticmethod
    def field(field_name, message: str, validate: ValidateField) -> FieldValidator:
        return FieldValidator(field_name, message, validate)

    @staticmethod
    def record(
        message: str, validate: ValidateRecord, field_name: str = None
    ) -> RecordValidator:
        return RecordValidator(message, validate, field_name)

    async def validate(self, record: Record) -> ValidationResult:
        return await ValidatorInstance(self, record).validate()


class ValidatorInstance:
    """Holds results for a run of validation."""

    def __init__(self, validator: Validator, record: Record) -> None:
        self.validator = validator
        self.record = record
        self.validation_result = ValidationResult()

    async def validate(self) -> ValidationResult:
        """Executes all validators against the record until one fails."""
        for validator in self.validator.validators:
            if isinstance(validator, FieldValidator):
                await self.validate_field(validator)
            elif isinstance(validator, RecordValidator):
                await self.validate_record(validator)
            else:
                raise Exception("Invalid validator")

            if not self.validation_result:
                return self.validation_result

        return self.validation_result

    async def validate_field(self, validator: FieldValidator) -> None:
        """Validate a single field."""
        field_name = validator.field_name
        value = self.record.get(field_name, None)
        result = await validator.validate(value)

        if not result:
            self.validation_result.add_message(field_name, validator.message)

    async def validate_record(self, validator: RecordValidator) -> None:
        """Validate an entire record."""
        result = await validator.validate(self.record)

        if not result:
            if validator.field_name:
                self.validation_result.add_message(
                    validator.field_name, validator.message
                )
            else:
                self.validation_result.add_general_message(validator.message)


class Validators:
    """Contains validator methods."""

    @staticmethod
    def is_present(val: Any) -> bool:
        """Validate that a value is not None"""
        if val is None:
            return False
        return True

    @staticmethod
    def optional(validator: ValidateFieldSync) -> ValidateFieldSync:
        """Makes a validator ignore None values"""

        def checker(val: Any) -> bool:
            if val is None:
                return True
            return validator(val)

        return checker

    @staticmethod
    def any(validators: List[ValidateFieldSync]) -> ValidateFieldSync:
        """Validates a value matches at least one validator."""

        def checker(val: Any) -> bool:
            for validator in validators:
                if validator(val):
                    return True
            return False

        return checker

    @staticmethod
    def is_type(typ: Type) -> ValidateFieldSync:
        """Validates that a value is of a type."""

        def checker(val: Any) -> bool:
            return isinstance(val, typ)

        return checker

    @staticmethod
    def is_member(members: Sequence) -> ValidateFieldSync:
        """Validates that a value is a member of a sequence."""

        def checker(val: Any) -> bool:
            return val in members

        return checker

    @staticmethod
    def is_not_empty(val: Any) -> ValidateFieldSync:
        """Validates that a string is not empty."""

        if not isinstance(val, str):
            return False
        return bool(val)

    @staticmethod
    def is_match(pattern: str) -> ValidateFieldSync:
        """Validates that a value matches a regular expression"""

        def checker(val: Any) -> bool:
            if not isinstance(val, str):
                return False
            return bool(re.match(pattern, val))

        return checker

    @staticmethod
    def is_length(min_len: Optional[int], max_len: Optional[int]) -> ValidateFieldSync:
        """Validates that a string value is within certain length bounds"""

        def checker(val: Any) -> bool:
            if not isinstance(val, str):
                return False
            if min_len is not None and len(val) < min_len:
                return False
            if max_len is not None and max_len < len(val):
                return False
            return True

        return checker

    @staticmethod
    def is_greater(bound: Number) -> ValidateFieldSync:
        """Validates that a number is greater than a bound."""

        def checker(val: Any) -> bool:
            if not isinstance(val, NUMBER_TYPES):
                return False
            return val > bound

        return checker

    @staticmethod
    def is_lesser(bound: Number) -> ValidateFieldSync:
        """Validates that a number is greater than a bound."""

        def checker(val: Any) -> bool:
            if not isinstance(val, NUMBER_TYPES):
                return False
            return val < bound

        return checker

    @staticmethod
    def is_greater_or_equal(bound: Number) -> ValidateFieldSync:
        """Validates that a number is greater than a bound."""

        def checker(val: Any) -> bool:
            if not isinstance(val, NUMBER_TYPES):
                return False
            return val >= bound

        return checker

    @staticmethod
    def is_lesser_or_equal(bound: Number) -> ValidateFieldSync:
        """Validates that a number is greater than a bound."""

        def checker(val: Any) -> bool:
            if not isinstance(val, NUMBER_TYPES):
                return False
            return val <= bound

        return checker

    @staticmethod
    def is_between(lbound: Number, ubound: Number) -> ValidateFieldSync:
        """Validates that a number is between two other numbers, inclusive."""

        def checker(val: Any) -> bool:
            if not isinstance(val, NUMBER_TYPES):
                return False
            return lbound <= val <= ubound

        return checker



"""Contains the StructuredTransaction class."""
# pylint: disable=unused-argument
from typing import Optional

from .data import TransactionData, TransactionInput, TransactionState
from .paging import Paging
from .policy_result import PolicyResult
from .transaction_result import TransactionResult
from .validation_result import ValidationResult


class TransactionBase:
    """Allows child classes to define parts of a standard transaction life-cycle:
    - initialize
    - authenticate
    - find
    - authorize
    - validate
    - perform
    """

    async def execute(
        self, input_data: TransactionInput, paging: Optional[Paging] = None
    ) -> TransactionResult:
        """Calls and response properly to overridden methods to run a standard API request."""
        state = await self.initialize(input_data)
        data: TransactionData = TransactionData(
            input_data, state, paging or Paging.default()
        )

        is_authenticated = await self.authenticate(data)
        if not is_authenticated:
            return TransactionResult.not_authenticated()

        missing_entity_name = await self.find(data)
        if missing_entity_name:
            return TransactionResult.not_found(missing_entity_name)

        policy_result = await self.authorize(data)
        if not policy_result.is_valid:
            return TransactionResult.not_authorized(policy_result)

        validation_result = await self.validate(data)
        if not validation_result.is_valid:
            return TransactionResult.not_valid(validation_result)

        result = await self.perform(data)
        return result

    async def __call__(self, input_data: TransactionInput) -> TransactionResult:
        return await self.execute(input_data)

    async def initialize(self, input_data: TransactionInput) -> TransactionState:
        return {}

    async def authenticate(self, data: TransactionData) -> bool:
        return True

    async def find(self, data: TransactionData) -> Optional[str]:
        return None

    async def authorize(self, data: TransactionData) -> PolicyResult:
        return PolicyResult.success()

    async def validate(self, data: TransactionData) -> ValidationResult:
        return ValidationResult.success()

    async def perform(self, data: TransactionData) -> TransactionResult:
        return TransactionResult.success()



"""Contains the TransactionCode Enum."""
from enum import Enum


class TransactionCode(Enum):
    """This enum represents all the status codes that can be returned in a TransactionResult."""

    success = 100
    found = 101
    created = 102
    updated = 103
    not_changed = 104
    deleted = 105
    failed = 200
    not_found = 201
    not_authenticated = 202
    not_authorized = 203
    not_valid = 204



"""Contains the transaction Data class."""
from dataclasses import dataclass
from typing import Any, Mapping, MutableMapping

from .paging import Paging

TransactionInput = Mapping[str, Any]
TransactionState = MutableMapping[str, Any]


@dataclass(frozen=True)
class TransactionData:
    """Contains input and state data for an API Transaction."""

    input: TransactionInput
    state: TransactionState
    paging: Paging



"""Contains the PolicyResult data structure."""
from typing import Optional

from .messages import MessageData, Messages


class PolicyResult:
    """Contains information about whether a policy passes or fails."""

    def __init__(self, is_valid: bool, messages: Optional[Messages]) -> None:
        self.is_valid: bool = is_valid
        self._messages: Optional[Messages] = messages

    def __bool__(self) -> bool:
        return self.is_valid

    @property
    def messages(self) -> MessageData:
        if self._messages is None:
            return {}
        return self._messages.messages

    @classmethod
    def success(cls) -> "PolicyResult":
        return PolicyResult(True, None)

    @classmethod
    def failed(cls, message: str) -> "PolicyResult":
        return PolicyResult(False, Messages().add_general_message(message))



"""Contains the Paging data structure."""
import math
from typing import Optional

from ntier.constants import PAGE_DEFAULT, PER_PAGE_DEFAULT, PER_PAGE_MAX


class Paging:
    """Contains paging information."""

    def __init__(self, page: int, per_page: int, total_records: int = 0):
        if page < 0:
            raise ValueError("page must be 0 or greater")
        if per_page < 1:
            raise ValueError("per_page must be 1 or greater")
        if per_page > PER_PAGE_DEFAULT:
            raise ValueError(f"per_page must be no greater than {PER_PAGE_MAX}")
        if total_records < 0:
            raise ValueError("total_records must be 0 or greater")

        self.page = page
        self.per_page = per_page
        self.total_records = total_records

    @classmethod
    def default(cls) -> "Paging":
        """Builds a default object."""
        return cls(PAGE_DEFAULT, PER_PAGE_DEFAULT)

    @classmethod
    def parse_paging(
        cls, raw_page: Optional[str], raw_per_page: Optional[str]
    ) -> "Paging":
        """Tries to parse paging from string values.

        Paging parameters often come in on the request as strings and need to be parsed as
        integers.
        """
        page: Optional[int] = None
        per_page: Optional[int] = None

        if raw_page:
            try:
                page = int(raw_page)
            except ValueError:
                page = PAGE_DEFAULT
        else:
            page = PAGE_DEFAULT
        if raw_per_page:
            try:
                per_page = int(raw_per_page)
            except ValueError:
                per_page = PER_PAGE_DEFAULT
        else:
            per_page = PER_PAGE_DEFAULT

        return cls(page, per_page)

    @property
    def limit(self) -> int:
        """Returns the limit based on the paging."""
        return self.per_page

    @property
    def offset(self) -> int:
        """Returns the offset based on the paging."""
        return (self.page - 1) * self.per_page

    @property
    def total_pages(self) -> int:
        """Return the total number of pages."""
        if self.total_records == 0:
            return 0
        return math.ceil(self.total_records / self.per_page)

    def set_total_records(self, value: int) -> "Paging":
        """Sets the total records."""
        if value < 0:
            raise ValueError("total_records must be 0 or greater")
        self.total_records = value
        return self



"""Contains the TransactionResult data structure."""
from typing import Any, Optional

from ntier.messages import Messages
from ntier.paging import Paging
from ntier.policy_result import PolicyResult
from ntier.transaction_code import TransactionCode
from ntier.validation_result import ValidationResult


class TransactionResult:
    """Represents a standard transaction result."""

    def __init__(
        self, is_valid: bool, status_code: TransactionCode, payload: Any
    ) -> None:
        self.is_valid: bool = is_valid
        self.status_code: TransactionCode = status_code
        self.payload: Any = payload
        self._paging: Optional[Paging] = None

    def __repr__(self):
        return "{}(is_valid={}, status_code={}, payload={}, paging={})".format(
            self.__class__.__name__,
            self.is_valid,
            self.status_code,
            self.payload,
            self._paging,
        )

    @property
    def paging(self) -> Paging:
        if self._paging is None:
            raise Exception("Paging is not set.")
        return self._paging

    @property
    def has_paging(self) -> bool:
        return self._paging is not None

    def set_paging(self, paging: Paging) -> "TransactionResult":
        self._paging = paging
        return self

    @classmethod
    def success(cls, payload: Any = None) -> "TransactionResult":
        return cls(True, TransactionCode.success, payload)

    @classmethod
    def found(cls, payload: Any = None) -> "TransactionResult":
        return cls(True, TransactionCode.found, payload)

    @classmethod
    def created(cls, payload: Any = None) -> "TransactionResult":
        return cls(True, TransactionCode.created, payload)

    @classmethod
    def updated(cls, payload: Any = None) -> "TransactionResult":
        return cls(True, TransactionCode.updated, payload)

    @classmethod
    def not_changed(cls, payload: Any = None) -> "TransactionResult":
        return cls(True, TransactionCode.not_changed, payload)

    @classmethod
    def deleted(cls, payload: Any = None) -> "TransactionResult":
        return cls(True, TransactionCode.deleted, payload)

    @classmethod
    def failed(cls, payload: Any = None) -> "TransactionResult":
        return cls(False, TransactionCode.failed, payload)

    @classmethod
    def not_found(cls, subject: str) -> "TransactionResult":
        msg = Messages().add_general_message(f"{subject} not found")
        return cls(False, TransactionCode.not_found, msg.messages)

    @classmethod
    def not_authenticated(cls) -> "TransactionResult":
        return cls(False, TransactionCode.not_authenticated, None)

    @classmethod
    def not_authorized(cls, policy_result: PolicyResult) -> "TransactionResult":
        return cls(False, TransactionCode.not_authorized, policy_result.messages)

    @classmethod
    def not_valid(cls, validation_result: ValidationResult) -> "TransactionResult":
        return cls(False, TransactionCode.not_valid, validation_result.messages)



"""Contains the ValidationResult data structure."""
from typing import Optional

from .messages import MessageData, Messages


class ValidationResult:
    """Contains information about whether a validation passes or fails."""

    def __init__(self, messages: Optional[Messages] = None) -> None:
        self._messages: Optional[Messages] = messages

    def __bool__(self) -> bool:
        return self.is_valid

    @property
    def is_valid(self):
        return not bool(self.messages)

    @property
    def messages(self) -> MessageData:
        if self._messages is None:
            return {}
        return self._messages.messages

    def add_message(self, subject: str, message: str) -> "ValidationResult":
        if self._messages is None:
            self._messages = Messages()
        self._messages.add_message(subject, message)
        return self

    def add_general_message(self, message: str) -> "ValidationResult":
        if self._messages is None:
            self._messages = Messages()
        self._messages.add_general_message(message)
        return self

    @classmethod
    def success(cls) -> "ValidationResult":
        return ValidationResult()

    @classmethod
    def failed(cls, messages: Messages) -> "ValidationResult":
        return ValidationResult(messages)

    @classmethod
    def builder(cls) -> "ValidationResult":
        return ValidationResult()



"""ntier API."""
from ntier.data import TransactionData, TransactionInput, TransactionState
from ntier.deserialize import Deserialize
from ntier.messages import Messages
from ntier.paging import Paging
from ntier.policy_result import PolicyResult
from ntier.serialize import Serialize
from ntier.transaction_base import TransactionBase
from ntier.transaction_code import TransactionCode
from ntier.transaction_result import TransactionResult
from ntier.validation_result import ValidationResult
from ntier.validator import Validator, Validators



"""Setup for package."""
import pathlib

from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
VERSION = "2.0.0"

setup(
    name="ntier",
    version=VERSION,
    description=(
        "A library containing data structures and classes used in making the domain layer "
        "of n-tier web applications"
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    author="Trey Cucco",
    author_email="fcucco@gmail.com",
    url="https://gitlab.com/tcucco/ntier",
    download_url="https://gitlab.com/tcucco/ntier/-/archive/master/ntier-master.tar.gz",
    package_data={"ntier": ["py.typed"]},
    packages=["ntier"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
    ],
    license="MIT",
    platforms="any",
    zip_safe=False,
    install_requires=["bleach~=3.2.1"],
)



"""Data deserialization methods."""
import base64
import datetime
import decimal
import uuid
from typing import Any, Callable, List, Mapping, NamedTuple, Optional, Union

from .deserialize_html import (HtmlAttributes, HtmlStripper, HtmlTags,
                               default_html_stripper)

Map = Callable[[Any], Any]
Record = Mapping[str, Any]


class MapDef(NamedTuple):
    extract: Map
    map: Map
    key: str


class Deserialize:
    """Contains deserialization methods. They take API data and turn it into domain data."""

    html_stripper: HtmlStripper = default_html_stripper

    @staticmethod
    def map_def(extract: Map, map: Map, key: str) -> MapDef:
        return MapDef(extract, map, key)

    @classmethod
    def register_html_stripper(cls, html_stripper: HtmlStripper) -> None:
        cls.html_stripper = html_stripper

    @classmethod
    def key(
        cls, key: str, map: Map, output_key: Optional[str] = None, fallback: Any = None
    ) -> MapDef:
        """Create a MapDef that pulls from a top-level key on a Mapping."""

        def extract(data: Record) -> Any:
            return data.get(key, fallback)

        return cls.map_def(extract, map, output_key or key)

    @classmethod
    def path(
        cls, keys: List[str], map: Map, output_key: str, fallback: Any = None
    ) -> MapDef:
        """Create a MapDef that pulls from a path in nested Mappings."""

        def extract(data: Record) -> Any:
            parts = iter(keys)
            value = data
            while True:
                try:
                    key = next(parts)
                except StopIteration:
                    return value
                if key in value:
                    value = value[key]
                else:
                    return fallback

        return cls.map_def(extract, map, output_key)

    @classmethod
    def map(
        cls, map_defs: Union[List[MapDef], Mapping[str, Map]]
    ) -> Callable[[Any], Record]:
        """Retrurns a function that will map any input into a mapping."""

        def mapper(data: Any) -> Record:
            if isinstance(map_defs, list):
                return {d.key: d.map(d.extract(data)) for d in map_defs}
            return {k: v(data.get(k)) for (k, v) in map_defs.items()}

        return mapper

    @classmethod
    def list(cls, fn: Map) -> Callable[[List], List]:
        """Returns a function that will apply a map to all items in a list."""

        def mapper(data: List) -> List:
            try:
                return [fn(val) for val in data]
            except TypeError:
                return []

        return mapper

    @classmethod
    def int(cls, val: Any) -> Optional[int]:
        """Try to parse a value as an integer."""
        try:
            return int(val)
        except (ValueError, TypeError):
            return None

    @classmethod
    def float(cls, val: Any) -> Optional[float]:
        """Try to parse a value as a float."""
        try:
            return float(val)
        except (ValueError, TypeError):
            return None

    @classmethod
    def date(cls, val: Any) -> Optional[datetime.date]:
        """Try to parse a value as an ISO date."""
        try:
            return datetime.date.fromisoformat(val)
        except (ValueError, TypeError):
            return None

    @classmethod
    def datetime(cls, val: Any) -> Optional[datetime.datetime]:
        """Try to parse a value as an ISO datetime."""
        try:
            return datetime.datetime.fromisoformat(val)
        except (ValueError, TypeError):
            return None

    @classmethod
    def uuid(cls, val: Any) -> Optional[uuid.UUID]:
        """Try to parse a value as an UUID."""
        try:
            return uuid.UUID(val)
        except (ValueError, TypeError):
            return None

    @classmethod
    def decimal(cls, val: Any) -> Optional[decimal.Decimal]:
        """Try to parse a decimal."""
        try:
            return decimal.Decimal(val)
        except (decimal.InvalidOperation, TypeError, ValueError):
            return None

    @classmethod
    def _text(
        cls, tags: HtmlTags, attributes: HtmlAttributes, val: Optional[str]
    ) -> Optional[str]:
        if val is None:
            return None

        val = str(val)
        if val is None:
            return None

        val = val.strip()
        if not val:
            return None

        val = cls.html_stripper(tags, attributes, val)

        return val

    @classmethod
    def text_with_html(
        cls, *, tags: HtmlTags = None, attributes: HtmlAttributes = None
    ) -> Callable[[Optional[str]], Optional[str]]:
        """Maps a value to a string."""

        def mapper(val: Optional[str]) -> Optional[str]:
            return cls._text(tags or [], attributes or {}, val)

        return mapper

    @classmethod
    def text(cls, val: Any) -> Optional[str]:
        return cls._text([], {}, val)

    @classmethod
    def base64(cls, val: Any) -> Optional[bytes]:
        """Try to parse base64 to bytes."""
        try:
            val = base64.b64decode(val, validate=True)
            if not val:
                return None
            return val
        except (TypeError, ValueError):
            return None
